#!/usr/bin/env python3
# given a list of integer variables and their values, this tool aims at finding which varibales can be obtained by the sum of others
# note : this tool is used to find the sums in the variables output by /proc/meminfo
from typing import Dict, List, Set, Tuple
import abc
from pathlib import Path
import json
import subprocess

VarName = str
VarValue = int
Sums = List[Dict[VarName, List[VarName]]]


class IntVariable():
    name: str
    value: int

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return f'{self.name}: {self.value}'


class ISumHandler(abc.ABC):

    @abc.abstractmethod
    def on_sum_found(self, total: IntVariable, components: List[IntVariable]):
        pass

    @abc.abstractmethod
    def on_end(self):
        pass


class SumPrinter(ISumHandler):

    def on_sum_found(self, total: IntVariable, components: List[IntVariable]):
        print(f"{total} = sum({components})")

    def on_end(self):
        pass


class SumExporter(ISumHandler):
    '''handler that exports the sums as a graphviz dot file
    '''
    sum_file_path: Path
    variables: Set[IntVariable]
    sums: Sums

    def __init__(self, sum_file_path: Path):
        self.sum_file_path = sum_file_path
        self.variables = set()
        self.sums = []

    def on_sum_found(self, total: IntVariable, components: List[IntVariable]):
        self.variables.add(total)
        for component in components:
            self.variables.add(component)
        print(f"sum found: {total.name} = {' + '.join([comp.name for comp in components])}")
        self.sums.append({'total': total.name, 'components': [comp.name for comp in components]})

    def on_end(self):
        print(f'exporting sums to {self.sum_file_path}')
        root_dict = {}
        root_dict['format'] = 'sums-v001'
        root_dict['variables'] = {var.name: var.value for var in self.variables}
        root_dict['sums'] = self.sums
        self.sum_file_path.write_text(json.dumps(root_dict))


def load_sums(sums_file_path: Path) -> Sums:
    with open(sums_file_path, 'rt', encoding='utf8') as sums_file:
        root_dict = json.loads(sums_file.read())
    assert root_dict['format'] == 'sums-v001'
    return root_dict['sums']


def explore(total: IntVariable, components: List[IntVariable], cur_sum: VarValue, cur_var_index: int, contrib_components: List[IntVariable], sum_handler: ISumHandler):
    # for i in range(cur_var_index):
    #     print('  ', end='')
    # print(f"exploring: cur_sum={cur_sum}, contrib_components={contrib_components}, cur_var={components[cur_var_index]}")
    for cur_var_presence in [False, True]:
        new_contrib_components = contrib_components.copy()
        new_sum = cur_sum
        if cur_var_presence:
            new_sum += components[cur_var_index].value
            new_contrib_components.append(components[cur_var_index])
        # print(f"new_sum={new_sum}")
        if new_sum == total.value:
            sum_handler.on_sum_found(total, new_contrib_components)
        elif new_sum < total.value and cur_var_index < len(components) - 1:
            explore(total, components, new_sum, cur_var_index + 1, new_contrib_components, sum_handler)


def find_numbers_summing_to(total: IntVariable, components: List[IntVariable], sum_handler: ISumHandler):
    # print(f'finding which combinations of {components} sum to {total}')
    explore(total, components, cur_sum=0, cur_var_index=0, contrib_components=[], sum_handler=sum_handler)


def find_sums(variables: Dict[VarName, VarValue], sum_handler: ISumHandler):
    int_vars = [IntVariable(k, v) for k, v in variables.items()]
    int_vars.sort(key=lambda var: var.value, reverse=True)
    print(f'set of integer variables to explore: {int_vars}')
    for total_index in range(len(int_vars) - 1):
        find_numbers_summing_to(int_vars[total_index], int_vars[total_index + 1:], sum_handler)
    sum_handler.on_end()


def sums_to_dot(sum_file_path: Path, dot_file_path: Path):

    def to_dot_symbol(string: str) -> str:
        return string.replace('(', '_').replace(')', '_').replace('-', '_')

    # from https://graphviz.org/doc/info/colors.html
    svg_colors = [
        "aliceblue",
        "antiquewhite",
        "aqua",
        "aquamarine",
        "azure",
        "beige",
        "bisque",
        "black",
        "blanchedalmond",
        "blue",
        "blueviolet",
        "brown",
        "burlywood",
        "cadetblue",
        "chartreuse",
        "chocolate",
        "coral",
        "cornflowerblue",
        "cornsilk",
        "crimson",
        "cyan",
        "darkblue",
        "darkcyan",
        "darkgoldenrod",
        "darkgray",
        "darkgreen",
        "darkgrey",
        "darkkhaki",
        "darkmagenta",
        "darkolivegreen",
        "darkorange",
        "darkorchid",
        "darkred",
        "darksalmon",
        "darkseagreen",
        "darkslateblue",
        "darkslategray",
        "darkslategrey",
        "darkturquoise",
        "darkviolet",
        "deeppink",
        "deepskyblue",
        "dimgray",
        "dimgrey",
        "dodgerblue",
        "firebrick",
        "floralwhite",
        "forestgreen",
        "fuchsia",
        "gainsboro",
        "ghostwhite",
        "gold",
        "goldenrod",
        "gray",
        "grey",
        "green",
        "greenyellow",
        "honeydew",
        "hotpink",
        "indianred",
        "indigo",
        "ivory",
        "khaki",
        "lavender",
        "lavenderblush",
        "lawngreen",
        "lemonchiffon",
        "lightblue",
        "lightcoral",
        "lightcyan",
        "lightgoldenrodyellow",
        "lightgray",
        "lightgreen",
        "lightgrey",
        "lightpink",
        "lightsalmon",
        "lightseagreen",
        "lightskyblue",
        "lightslategray",
        "lightslategrey",
        "lightsteelblue",
        "lightyellow",
        "lime",
        "limegreen",
        "linen",
        "magenta",
        "maroon",
        "mediumaquamarine",
        "mediumblue",
        "mediumorchid",
        "mediumpurple",
        "mediumseagreen",
        "mediumslateblue",
        "mediumspringgreen",
        "mediumturquoise",
        "mediumvioletred",
        "midnightblue",
        "mintcream",
        "mistyrose",
        "moccasin",
        "navajowhite",
        "navy",
        "oldlace",
        "olive",
        "olivedrab",
        "orange",
        "orangered",
        "orchid",
        "palegoldenrod",
        "palegreen",
        "paleturquoise",
        "palevioletred",
        "papayawhip",
        "peachpuff",
        "peru",
        "pink",
        "plum",
        "powderblue",
        "purple",
        "red",
        "rosybrown",
        "royalblue",
        "saddlebrown",
        "salmon",
        "sandybrown",
        "seagreen",
        "seashell",
        "sienna",
        "silver",
        "skyblue",
        "slateblue",
        "slategray",
        "slategrey",
        "snow",
        "springgreen",
        "steelblue",
        "tan",
        "teal",
        "thistle",
        "tomato",
        "turquoise",
        "violet",
        "wheat",
        "white",
        "whitesmoke",
        "yellow",
        "yellowgreen"]

    print(f'creating {dot_file_path} from {sum_file_path}')
    with open(sum_file_path, 'rt', encoding='utf8') as sum_file:
        root_dict = json.load(sum_file)
    # print(root_dict)
    variables = {IntVariable(var_name, var_value) for var_name, var_value in root_dict['variables'].items()}
    # print(variables)
    sums = root_dict['sums']
    # print(sums)
    with open(dot_file_path, 'wt+', encoding='utf8') as dot_file:
        dot_file.write(f'digraph {to_dot_symbol(sum_file_path.stem)}\n')
        dot_file.write('{\n')
        dot_file.write('layout = neato\n')
        dot_file.write('overlap = false\n')

        for variable in variables:
            dot_file.write(f'{to_dot_symbol(variable.name)} [label="{variable.name}={variable.value}"];\n')
        sum_index = 0
        for s in sums:
            dot_file.write(f"{to_dot_symbol(s['total'])} -> {{{','.join([to_dot_symbol(comp) for comp in s['components']])}}} [color={svg_colors[sum_index % len(svg_colors)]}, label={sum_index}];\n")
            sum_index += 1
        dot_file.write('}\n')


def find_and_graph_sums(variables: Dict[str, int], set_id: str):
    # set_id: eg 'set001' 'alambix97-meminfo'
    sums_file_path = Path(f'{set_id}-sums.json')
    dot_file_path = Path(f'{set_id}-sums.dot')
    svg_file_path = Path(f'{set_id}-sums.svg')
    sum_handler = SumExporter(sums_file_path)
    find_sums(variables, sum_handler)
    sums_to_dot(sums_file_path, dot_file_path)
    print(f'creating {svg_file_path} from {dot_file_path}')
    subprocess.run(f'dot -Tsvg {dot_file_path} > {svg_file_path}', shell=True, check=True)


def remove_invalid_sums(input_sums: Sums, variables: Dict[VarName, VarValue]) -> Tuple[Sums, Sums]:
    """
    returns the tuple valid_sums, invalid_sums after checking each of the sum of input_sum againts the given variables
    """
    valid_sums: Sums = []
    invalid_sums: Sums = []
    for _sum in input_sums:
        # for total_var_id, components in input_sums.items():
        total_var_id = _sum['total']
        components = _sum['components']

        computed_total_value = 0
        for component_var_id in components:
            computed_total_value += variables[component_var_id]
        real_total_value = variables[total_var_id]
        if computed_total_value == real_total_value:
            valid_sums.append(_sum)
        else:
            invalid_sums.append(_sum)
    return (valid_sums, invalid_sums)
